/* pam/pam_skey.c - A module to interface the Linux-PAM Pluggable
 * Authentication Module to the S/Key one-time-password system.
 *
 * This compiles into pam_skey.so which should be installed into
 * /lib[<qual>]/security/pam_skey.so.  It can be used through a pam.conf
 * line, preferably stacked with pam_unix, similar to 
 * telnet     auth       required     pam_unix.so
 * telnet     auth       required     pam_skey.so
 *
 * This module provides only the 'auth' capability.
 *
 * You need both the skey and Linux-PAM packages for this to be
 * useful.
 *
 * Copyright (C) 1997 Pharos Intellectual Property Pty Ltd.
 * You may redistribute this module under the GNU Public License
 * version 2. 
 *
 * Author: Martin Pool <m.pool@pharos.com.au>
 * USE PGP TO VERIFY THIS PACKAGE.
 */

/* NO WARRANTY: THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  */

/* The following options are permitted in the pam.conf file:
   
   debug = enable debugging output to syslog(AUTHPRIV.DEBUG)

   noecho = don't echo the password back.  This is not on by default
   because S/Key passwords are long, and because the system design assumes
   that they may be snooped anyhow.  Turn it on if you wish.

   use_first_pass and try_first_pass are not implemented because it is
   unlikely that the S/Key password would be the same as that of
   another PAM.
   
 */
#include "config.h"

#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <security/skey.h>

/* Capabilities provided by this module */
#define PAM_SM_AUTH

#include <security/pam_modules.h>

#define DEBUG	1
#define AUDIT	2
#define NOECHO	4

/* Read "command-line" options */
static int read_options(int argc, const char **argv, int *ctrl)
{
	int i;
	for (i = 0; i < argc; i++) {
		if (!strncmp(argv[i], "debug", 5))
			*ctrl |= DEBUG;
		else if (!strncmp(argv[i], "audit", 5))
			*ctrl |= AUDIT;
		else if (!strncmp(argv[i], "noecho", 6))
			*ctrl |= NOECHO;
		else		/* PAM spec says unknown options must be ignored */
			syslog(LOG_AUTHPRIV | LOG_WARNING,
			       "pam_skey: unknown argument '%s'", argv[i]);
	}
	return PAM_SUCCESS;
}

/*
 * The guts of PAM-S/Key authentication.  Returns a PAM_* status code,
 * depending on whether the user was successfully authenticated or not.
 * This fn attempts to map the error codes returned by S/Key into
 * the corresponding PAM codes: they ought to be close but probably not perfect.
 */
static int _pam_auth_skey(pam_handle_t * pamh, int flags, int ctrl)
{
	/* Unlike the skey-provided login(1), this module never accepts the
	 * unix passwd - use the pam.conf to stack pam_skey and
	 * pam_unix if you want to give the user the option. */
	char const *user;
	char challenge[1024];
	int rc;
	struct skey sk;
	struct pam_conv *conv;
	struct pam_message msg[1];
	const struct pam_message *pmsg[1];
	struct pam_response *presponse = NULL;

	/* Get the users name, if it is not already known */
	rc = pam_get_user(pamh, &user, "s/key login: ");
	if (rc != PAM_SUCCESS)
		return rc;
	if (ctrl & DEBUG)
		syslog(LOG_AUTHPRIV | LOG_NOTICE,
		       "pam_skey login attempt for '%s'", user);

	/* Get the string with which to challenge the user */
	rc = skeychallenge(&sk, (char *) user, challenge);
	switch (rc) {
	case -1:
		syslog(LOG_AUTHPRIV | LOG_WARNING,
		       "pam_skey: system error finding challenge string");
		return PAM_SYSTEM_ERR;
	case 1:
		syslog(LOG_AUTHPRIV | LOG_WARNING,
		       "pam_skey: user '%s' unknown", user);
		return PAM_USER_UNKNOWN;
	}
	if (ctrl & DEBUG)
		syslog(LOG_AUTHPRIV | LOG_NOTICE,
		       "pam_skey login challenge is '%s'", challenge);
	strcat(challenge, ": ");

	/* Get the conversation function with which to prompt the user for
	 * their password. */
	rc = pam_get_item(pamh, PAM_CONV, (const void **) &conv);
	if (rc != PAM_SUCCESS)
		return rc;

	/* Prompt the user with the challenge, and get their response */
	pmsg[0] = &msg[0];
	msg[0].msg_style =
	    (ctrl & NOECHO) ? PAM_PROMPT_ECHO_OFF : PAM_PROMPT_ECHO_ON;
	msg[0].msg = challenge;
	presponse = NULL;

	rc = conv->conv(1, (const struct pam_message **) &pmsg,
			&presponse, conv->appdata_ptr);
	if (rc != PAM_SUCCESS)
		return rc;

	/* a bit insecure */
	if (ctrl & AUDIT)
		syslog(LOG_AUTHPRIV | LOG_NOTICE,
		       "pam_skey response is '%s'", presponse->resp);

	/* Check the response with S/Key.  S/Key uses the username specified
	 * above. */
	rc = skeyverify(&sk, presponse->resp);
	switch (rc) {
	case -1:
		syslog(LOG_AUTHPRIV | LOG_WARNING,
		       "pam_skey: S/Key system error or authentication failure");
		return PAM_AUTH_ERR;
	case 1:
		syslog(LOG_AUTHPRIV | LOG_WARNING,
		       "pam_skey: user '%s' S/Key authentication failed",
		       user);
		return PAM_AUTH_ERR;
	}
	if (ctrl & DEBUG)
		syslog(LOG_AUTHPRIV | LOG_NOTICE, "pam_skey login OK");

	return PAM_SUCCESS;
}

/* pam_sm_authenticate: Entry point for a PAM request to validate a user. */
PAM_EXTERN int pam_sm_authenticate(pam_handle_t * pamh,
				   int flags, int argc, const char **argv)
{
	int rc, ctrl;
	rc = read_options(argc, argv, &ctrl);
	if (rc != PAM_SUCCESS)
		return rc;
	return _pam_auth_skey(pamh, flags, ctrl);
}

/* Is there anything this can usefully do?  What are credentials, anyhow? */
PAM_EXTERN int pam_sm_setcred(pam_handle_t * pamh,
			      int flags, int argc, const char **argv)
{
	return PAM_SUCCESS;
}


#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_skey_modstruct = {
	"pam_skey",
	pam_sm_authenticate,
	pam_sm_setcred,
	NULL,
	NULL,
	NULL,
	NULL,
};

#endif

/* end of module definition */
